/*
   Copyright (c) 2006, 2007, 2008 Z RESEARCH, Inc. <http://www.zresearch.com>
   This file is part of GlusterFS.

   GlusterFS is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published
   by the Free Software Foundation; either version 3 of the License,
   or (at your option) any later version.

   GlusterFS is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see
   <http://www.gnu.org/licenses/>.
*/

/*TODO: check for non null wb_file_data before getting wb_file */


#ifndef _CONFIG_H
#define _CONFIG_H
#include "config.h"
#endif

#include "glusterfs.h"
#include "logging.h"
#include "dict.h"
#include "xlator.h"
#include "list.h"
#include "compat.h"
#include "compat-errno.h"
#include "common-utils.h"

#define MAX_VECTOR_COUNT 8
 
typedef struct list_head list_head_t;
struct wb_conf;
struct wb_page;
struct wb_file;


struct wb_conf {
  size_t aggregate_size;
  size_t window_size;
  char flush_behind;
};


typedef struct wb_local {
  list_head_t winds;
  struct wb_file *file;
} wb_local_t;


typedef struct write_request {
  call_frame_t *frame;
  off_t offset;
  struct iovec *vector;
  int32_t count;
  dict_t *refs;
  char write_behind;
  char stack_wound;
  char got_reply;
  list_head_t list;
  list_head_t winds;
  list_head_t unwinds;
  gf_lock_t lock;
} wb_write_request_t;


struct wb_file {
  int disabled;
  int disable_till;
  off_t offset;
  size_t window_size;
  int32_t refcount;
  int32_t op_ret;
  int32_t op_errno;
  list_head_t request;
  fd_t *fd;
  gf_lock_t lock;
  xlator_t *this;
};


typedef struct wb_conf wb_conf_t;
typedef struct wb_page wb_page_t;
typedef struct wb_file wb_file_t;


int32_t 
wb_process_queue (call_frame_t *frame, wb_file_t *file, char flush_all);

int32_t
wb_sync (call_frame_t *frame, wb_file_t *file, list_head_t *winds);

int32_t
wb_sync_all (call_frame_t *frame, wb_file_t *file);

int32_t 
__wb_mark_winds (list_head_t *list, list_head_t *winds, size_t aggregate_size);

int32_t
wb_stack_wind (list_head_t *list, list_head_t *winds, char dont_aggregate);



wb_file_t *
wb_file_ref (wb_file_t *file)
{
  LOCK (&file->lock);
  {
    file->refcount++;
  }
  UNLOCK (&file->lock);

  return file;
}


void
wb_file_unref (wb_file_t *file)
{
  int32_t refcount;

  LOCK (&file->lock);
  {
    refcount = --file->refcount;
  }
  UNLOCK (&file->lock);

  if (!refcount)
    {
      LOCK_DESTROY (&file->lock);
      FREE (file);
    }

  return;
}


int32_t
wb_sync_cbk (call_frame_t *frame,
             void *cookie,
             xlator_t *this,
             int32_t op_ret,
             int32_t op_errno,
	     struct stat *stbuf)
{
  wb_local_t *local = NULL;
  list_head_t *winds = NULL;
  wb_file_t *file = NULL;
  wb_write_request_t *request = NULL, *dummy = NULL;
  char written_behind = 0;
  int32_t ret = 0;
  call_frame_t *unwind_frame = NULL;
  size_t size = 0;

  local = frame->local;
  winds = &local->winds;
  file = local->file;

  list_for_each_entry_safe (request, dummy, winds, winds)
    {
      unwind_frame = NULL;
      size = 0;

      LOCK (&request->lock);
      {
	request->got_reply = 1;
	written_behind = request->write_behind;

	if (!written_behind)
	  {
	    unwind_frame = request->frame;
	    size = iov_length (request->vector, request->count);
	  }
      }
      UNLOCK (&request->lock);

      if (!written_behind)
	{
	  ret = (op_ret > 0) ? size : op_ret;
	  STACK_UNWIND (unwind_frame, ret, op_errno, stbuf);
	} 
    }

  if (op_ret == -1)
    {
      file->op_ret = op_ret;
      file->op_errno = op_errno;
    }

  wb_process_queue (frame, file, 0); 

  STACK_DESTROY (frame->root);
  wb_file_unref (file);

  return 0;
}


int32_t
wb_sync_all (call_frame_t *frame, wb_file_t *file) 
{
  list_head_t winds;
  int32_t bytes = 0;

  INIT_LIST_HEAD (&winds);

  LOCK (&file->lock);
  {
    bytes = __wb_mark_winds (&file->request, &winds, 0);
  }
  UNLOCK (&file->lock);

  wb_sync (frame, file, &winds);

  return bytes;
}


int32_t
wb_sync (call_frame_t *frame, wb_file_t *file, list_head_t *winds)
{
  wb_write_request_t *dummy = NULL, *request = NULL, *first_request = NULL;
  size_t total_count = 0, count = 0;
  size_t copied = 0;
  call_frame_t *sync_frame = NULL;
  dict_t *refs = NULL;
  wb_local_t *local = NULL;
  struct iovec *vector = NULL;
  int32_t bytes = 0;
  size_t bytecount = 0;

  list_for_each_entry (request, winds, winds)
    {
      total_count += request->count;
      bytes += iov_length (request->vector, request->count);
    }

  if (!total_count) {
      return 0;
    }

  list_for_each_entry_safe (request, dummy, winds, winds)
    {
      if ((count != 0) && (count + request->count > MAX_VECTOR_COUNT)) {
	sync_frame = copy_frame (frame);  
	sync_frame->local = local;
	sync_frame->root->req_refs = dict_ref (refs);

	STACK_WIND (sync_frame,
		    wb_sync_cbk,
		    FIRST_CHILD(sync_frame->this),
		    FIRST_CHILD(sync_frame->this)->fops->writev,
		    file->fd, vector,
		    count, first_request->offset);
	
	dict_unref (refs);
	FREE (vector);
	first_request = NULL;
	refs = NULL;
	vector = NULL;
	copied = count = 0;
      }

      if (!vector) {
	vector = malloc (VECTORSIZE (MAX_VECTOR_COUNT));
	refs = get_new_dict ();
	refs->is_locked = 1;

	local = calloc (1, sizeof (*local));
	local->file = wb_file_ref (file);
	INIT_LIST_HEAD (&local->winds);

	first_request = request;
      }

      count += request->count;
      bytecount = VECTORSIZE (request->count);
      memcpy (((char *)vector)+copied,
	      request->vector,
	      bytecount);
      copied += bytecount;

      if (request->refs) {
	  dict_copy (request->refs, refs);
      }

      list_del_init (&request->winds);
      list_add_tail (&request->winds, &local->winds);
    }

  sync_frame = copy_frame (frame);  
  sync_frame->local = local;
  sync_frame->root->req_refs = dict_ref (refs);

  STACK_WIND (sync_frame,
	      wb_sync_cbk,
	      FIRST_CHILD(sync_frame->this),
	      FIRST_CHILD(sync_frame->this)->fops->writev,
	      file->fd, vector,
	      count, first_request->offset);

  dict_unref (refs);
  FREE (vector);

  return bytes;
}


int32_t 
wb_stat_cbk (call_frame_t *frame,
             void *cookie,
             xlator_t *this,
             int32_t op_ret,
             int32_t op_errno,
             struct stat *buf)
{
  wb_local_t *local = frame->local;

  if (local->file)
    {
      wb_file_unref (local->file);
    }

  STACK_UNWIND (frame, op_ret, op_errno, buf);

  return 0;
}


int32_t
wb_stat (call_frame_t *frame,
         xlator_t *this,
         loc_t *loc)
{
  wb_file_t *file = NULL;
  fd_t *iter_fd = NULL;
  wb_local_t *local = NULL;

  if (loc->inode)
    {
      LOCK (&(loc->inode->lock));
      {
	list_for_each_entry (iter_fd, &(loc->inode->fds), inode_list)
	  {
	    if (dict_get (iter_fd->ctx, this->name))
	      {
		file = data_to_ptr (dict_get (iter_fd->ctx, this->name));
		break;
	      }
	  }
      }
      UNLOCK (&(loc->inode->lock));

      if (file)
	{
	  wb_sync_all (frame, file);
	}
    }

  local = calloc (1, sizeof (*local));
  if (file)
    {
      local->file = wb_file_ref (file);
    }
  frame->local = local;

  STACK_WIND (frame, wb_stat_cbk,
              FIRST_CHILD(this),
              FIRST_CHILD(this)->fops->stat,
              loc);
  return 0;
}


int32_t 
wb_fstat (call_frame_t *frame,
          xlator_t *this,
          fd_t *fd)
{
  wb_file_t *file = NULL;
  wb_local_t *local = NULL;
  
  if (!dict_get (fd->ctx, this->name))
    {
      gf_log (this->name, GF_LOG_ERROR, "returning EBADFD");
      STACK_UNWIND (frame, -1, EBADFD, NULL);
      return 0;
    }

  file = data_to_ptr (dict_get (fd->ctx, this->name));

  if (file)
    {
      wb_sync_all (frame, file);
    }

  local = calloc (1, sizeof (*local));
  if (file)
    {
      local->file = wb_file_ref (file);
    }
  frame->local = local;

  STACK_WIND (frame,
              wb_stat_cbk,
              FIRST_CHILD(this),
              FIRST_CHILD(this)->fops->fstat,
              fd);
  return 0;
}


int32_t
wb_truncate_cbk (call_frame_t *frame,
                 void *cookie,
                 xlator_t *this,
                 int32_t op_ret,
                 int32_t op_errno,
                 struct stat *buf)
{
  wb_local_t *local = frame->local;

  if (local->file)
    {
      wb_file_unref (local->file);
    }

  STACK_UNWIND (frame, op_ret, op_errno, buf);
  return 0;
}


int32_t 
wb_truncate (call_frame_t *frame,
             xlator_t *this,
             loc_t *loc,
             off_t offset)
{
  wb_file_t *file = NULL;
  fd_t *iter_fd = NULL;
  wb_local_t *local = NULL;

  if (loc->inode)
    {
      LOCK (&(loc->inode->lock));
      {
	list_for_each_entry (iter_fd, &(loc->inode->fds), inode_list)
	  {
	    if (dict_get (iter_fd->ctx, this->name))
	      {
		file = data_to_ptr (dict_get (iter_fd->ctx, this->name));
		break;
	      }
	  }
      }
      UNLOCK (&(loc->inode->lock));
    
      if (file)
	{
	  wb_sync_all (frame, file);
	}
    }
  
  local = calloc (1, sizeof (*local));
  if (file)
    {
      local->file = wb_file_ref (file);
    }
  frame->local = local;

  STACK_WIND (frame,
              wb_truncate_cbk,
              FIRST_CHILD(this),
              FIRST_CHILD(this)->fops->truncate,
              loc,
              offset);
  return 0;
}


int32_t
wb_ftruncate (call_frame_t *frame,
              xlator_t *this,
              fd_t *fd,
              off_t offset)
{
  wb_file_t *file = NULL;
  wb_local_t *local = NULL;

  if (!dict_get (fd->ctx, this->name))
    {
      gf_log (this->name, GF_LOG_ERROR, "returning EBADFD");
      STACK_UNWIND (frame, -1, EBADFD, NULL);
      return 0;
    }

  file = data_to_ptr (dict_get (fd->ctx, this->name));

  if (file)
    {
      wb_sync_all (frame, file);
    }

  local = calloc (1, sizeof (*local));
  if (file)
    {
      local->file = wb_file_ref (file);
    }
  frame->local = local;

  STACK_WIND (frame,
              wb_truncate_cbk,
              FIRST_CHILD(this),
              FIRST_CHILD(this)->fops->ftruncate,
              fd,
              offset);
  return 0;
}


int32_t 
wb_utimens_cbk (call_frame_t *frame,
                void *cookie,
                xlator_t *this,
                int32_t op_ret,
                int32_t op_errno,
                struct stat *buf)
{
  wb_local_t *local = frame->local;

  if (local->file)
    {
      wb_file_unref (local->file);
    }

  STACK_UNWIND (frame, op_ret, op_errno, buf);
  return 0;
}


int32_t 
wb_utimens (call_frame_t *frame,
            xlator_t *this,
            loc_t *loc,
            struct timespec tv[2])
{
  wb_file_t *file = NULL;
  fd_t *iter_fd = NULL;
  wb_local_t *local = NULL;

  if (loc->inode)
    {
      LOCK (&(loc->inode->lock));
      {
	list_for_each_entry (iter_fd, &(loc->inode->fds), inode_list)
	  {
	    if (dict_get (iter_fd->ctx, this->name))
	      {
		file = data_to_ptr (dict_get (iter_fd->ctx, this->name));
		break;
	      }
	  }
      }
      UNLOCK (&(loc->inode->lock));

      if (file)
	{
	  wb_sync_all (frame, file);
	}
    }

  local = calloc (1, sizeof (*local));
  if (file)
    {
      local->file = wb_file_ref (file);
    }
  frame->local = local;

  STACK_WIND (frame,
              wb_utimens_cbk,
              FIRST_CHILD(this),
              FIRST_CHILD(this)->fops->utimens,
              loc,
              tv);
  return 0;
}

int32_t
wb_open_cbk (call_frame_t *frame,
             void *cookie,
             xlator_t *this,
             int32_t op_ret,
             int32_t op_errno,
             fd_t *fd)
{
  int32_t flags = 0;
  wb_file_t *file = NULL;

  if (op_ret != -1)
    {
      file = calloc (1, sizeof (*file));
      INIT_LIST_HEAD (&file->request);

      file->fd= fd;
      file->this = this;
      file->disable_till = 1048576;

      dict_set (fd->ctx, this->name,
		data_from_static_ptr (file));

      /* If mandatory locking has been enabled on this file,
	 we disable caching on it */

      if ((fd->inode->st_mode & S_ISGID) && !(fd->inode->st_mode & S_IXGRP))
	file->disabled = 1;

      /* If O_DIRECT then, we disable chaching */
      if (frame->local)
	{
	  flags = *((int32_t *)frame->local);
	  if (flags & O_DIRECT)
	    file->disabled = 1;
	}

      LOCK_INIT (&file->lock);
      wb_file_ref (file);
    }

  STACK_UNWIND (frame, op_ret, op_errno, fd);
  return 0;
}


int32_t
wb_open (call_frame_t *frame,
         xlator_t *this,
         loc_t *loc,
         int32_t flags,
	 fd_t *fd)
{
  frame->local = calloc (1, sizeof(int32_t));
  *((int32_t *)frame->local) = flags;

  STACK_WIND (frame,
              wb_open_cbk,
              FIRST_CHILD(this),
              FIRST_CHILD(this)->fops->open,
              loc, flags, fd);
  return 0;
}


int32_t
wb_create_cbk (call_frame_t *frame,
               void *cookie,
               xlator_t *this,
               int32_t op_ret,
               int32_t op_errno,
               fd_t *fd,
               inode_t *inode,
               struct stat *buf)
{
  wb_file_t *file = NULL;

  if (op_ret != -1)
    {
      file = calloc (1, sizeof (*file));
      INIT_LIST_HEAD (&file->request);

      file->fd= fd;
      file->disable_till = 1048576;
      file->this = this;

      dict_set (fd->ctx, this->name,
		data_from_static_ptr (file));

    /* 
     * If mandatory locking has been enabled on this file,
     * we disable caching on it
     */
      if ((fd->inode->st_mode & S_ISGID) && 
	  !(fd->inode->st_mode & S_IXGRP))
	{
	  file->disabled = 1;
	}

      LOCK_INIT (&file->lock);
      wb_file_ref (file);
    }

  STACK_UNWIND (frame, op_ret, op_errno, fd, inode, buf);
  return 0;
}


int32_t
wb_create (call_frame_t *frame,
           xlator_t *this,
	   loc_t *loc,
           int32_t flags,
           mode_t mode,
	   fd_t *fd)
{
  STACK_WIND (frame,
              wb_create_cbk,
              FIRST_CHILD(this),
              FIRST_CHILD(this)->fops->create,
	      loc, flags, mode, fd);
  return 0;
}


int32_t 
__wb_cleanup_queue (wb_file_t *file)
{
  wb_write_request_t *request = NULL, *dummy = NULL;
  int32_t bytes = 0;
  char got_reply = 0;

  list_for_each_entry_safe (request, dummy, &file->request, list)
    {
      LOCK (&request->lock);
      {
	got_reply = request->got_reply;
      }
      UNLOCK (&request->lock);

      if (got_reply)
	{
	  bytes += iov_length (request->vector, request->count);
	  list_del_init (&request->list);

	  FREE (request->vector);
	  dict_unref (request->refs);
	  LOCK_DESTROY (&request->lock);
      
	  FREE (request);
	}
    }

  return bytes;
}


int32_t 
__wb_mark_wind_all (list_head_t *list, list_head_t *winds)
{
  wb_write_request_t *request = NULL;
  size_t size = 0;

  list_for_each_entry (request, list, list)
    {
      LOCK (&request->lock);
      {
	if (!request->stack_wound)
	  {
	    size += iov_length (request->vector, request->count);
	    request->stack_wound = 1;
	    list_add_tail (&request->winds, winds);
	  }
      }
      UNLOCK (&request->lock);
    }

  return size;
}


size_t 
__wb_get_aggregate_size (list_head_t *list)
{
  wb_write_request_t *request = NULL;
  size_t size = 0;

  list_for_each_entry (request, list, list)
    {
      LOCK (&request->lock);
      {
	if (!request->stack_wound)
	  {
	    size += iov_length (request->vector, request->count);
	  }
      }
      UNLOCK (&request->lock);
    }

  return size;
}

uint32_t
__wb_get_incomplete_writes (list_head_t *list)
{
  wb_write_request_t *request = NULL;
  uint32_t count = 0;

  list_for_each_entry (request, list, list)
    {
      LOCK (&request->lock);
      {
	if (request->stack_wound && !request->got_reply)
	  {
	    count++;
	  }
      }
      UNLOCK (&request->lock);
    }

  return count;
}

int32_t
__wb_mark_winds (list_head_t *list, list_head_t *winds, size_t aggregate_conf)
{
  size_t aggregate_current = 0;
  uint32_t incomplete_writes = 0;

  incomplete_writes = __wb_get_incomplete_writes (list);

  aggregate_current = __wb_get_aggregate_size (list);

  if ((incomplete_writes == 0) || (aggregate_current >= aggregate_conf))
    {
      __wb_mark_wind_all (list, winds);
    }

  return aggregate_current;
}


size_t
__wb_get_window_size (list_head_t *list)
{
  wb_write_request_t *request = NULL;
  size_t size = 0;

  list_for_each_entry (request, list, list)
    {
      LOCK (&request->lock);
      {
	if (request->write_behind && !request->got_reply)
	  {
	    size += iov_length (request->vector, request->count);
	  }
      }
      UNLOCK (&request->lock);
    }

  return size;
}


size_t 
__wb_mark_unwind_till (list_head_t *list, list_head_t *unwinds, size_t size)
{
  size_t written_behind = 0;
  wb_write_request_t *request = NULL;

  list_for_each_entry (request, list, list)
    {
      if (size > written_behind)
	{
	  LOCK (&request->lock);
	  {
	    if (!request->write_behind)
	      {
		written_behind += iov_length (request->vector, request->count);
		request->write_behind = 1;
		list_add_tail (&request->unwinds, unwinds);
	      }
	  }
	  UNLOCK (&request->lock);
	}
      else
	{
	  break;
	}
    }

  if (written_behind < size)
    {
      //      wb_debug_function ();
    }

  return written_behind;
}


int32_t 
__wb_mark_unwinds (list_head_t *list, list_head_t *unwinds, size_t window_conf)
{
  size_t window_current = 0;

  window_current = __wb_get_window_size (list);
  if (window_current < window_conf)
    {
      window_current += __wb_mark_unwind_till (list, unwinds,
					       window_conf - window_current);
    }

  return window_current;
}


int32_t
wb_stack_unwind (list_head_t *unwinds)
{
  size_t size = 0;
  struct stat buf = {0,};
  wb_write_request_t *request = NULL, *dummy = NULL;

  list_for_each_entry_safe (request, dummy, unwinds, unwinds)
    {
      LOCK (&request->lock);
      {
	size = iov_length (request->vector, request->count);
	list_del_init (&request->unwinds);
      }
      UNLOCK (&request->lock);

      STACK_UNWIND (request->frame, size, 0, &buf);
    }

  return 0;
}


int32_t
wb_do_ops (call_frame_t *frame, wb_file_t *file, list_head_t *winds, list_head_t *unwinds)
{
  /* copy the frame before calling wb_stack_unwind, since this request containing current frame might get unwound */
  /*  call_frame_t *sync_frame = copy_frame (frame); */
 
  wb_stack_unwind (unwinds);
  wb_sync (frame, file, winds);

  return 0;
}


int32_t 
wb_process_queue (call_frame_t *frame, wb_file_t *file, char flush_all) 
{
  list_head_t winds, unwinds;
  size_t size = 0;
  wb_conf_t *conf = file->this->private;

  INIT_LIST_HEAD (&winds);
  INIT_LIST_HEAD (&unwinds);

  if (!file)
    {
      return -1;
    }

  size = flush_all ? 0 : conf->aggregate_size;
  LOCK (&file->lock);
  {
    __wb_cleanup_queue (file);
    __wb_mark_winds (&file->request, &winds, size);
    __wb_mark_unwinds (&file->request, &unwinds, conf->window_size);
  }
  UNLOCK (&file->lock);

  wb_do_ops (frame, file, &winds, &unwinds);
  return 0;
}


wb_write_request_t *
wb_enqueue (wb_file_t *file, 
	    call_frame_t *frame,
	    struct iovec *vector,
	    int32_t count,
	    off_t offset)
{
  wb_write_request_t *request = NULL;

  request = calloc (1, sizeof (*request));

  INIT_LIST_HEAD (&request->list);
  INIT_LIST_HEAD (&request->winds);
  INIT_LIST_HEAD (&request->unwinds);

  LOCK_INIT (&request->lock);

  request->frame = frame;
  request->vector = iov_dup (vector, count);
  request->count = count;
  request->offset = offset;
  request->refs = dict_ref (frame->root->req_refs);

  LOCK (&file->lock);
  {
    list_add_tail (&request->list, &file->request);
    file->offset = offset + iov_length (vector, count);
  }
  UNLOCK (&file->lock);

  return request;
}


int32_t
wb_writev (call_frame_t *frame,
           xlator_t *this,
           fd_t *fd,
           struct iovec *vector,
           int32_t count,
           off_t offset)
{
  wb_file_t *file = NULL;
  char offset_expected = 1;
  call_frame_t *process_frame = NULL;

  process_frame = copy_frame (frame);

  if (!dict_get (fd->ctx, this->name))
    {
      gf_log (this->name, GF_LOG_ERROR, "returning EBADFD");
      STACK_UNWIND (frame, -1, EBADFD, NULL);
      return 0;
    }

  file = data_to_ptr (dict_get (fd->ctx, this->name));

  if (!file)
    {
      gf_log (this->name, GF_LOG_ERROR,
	      "wb_file not found for fd %p", fd);
      STACK_UNWIND (frame, -1, EBADFD, NULL);
      return 0;
    }

  LOCK (&file->lock);
  {
    if (file->offset != offset)
      offset_expected = 0;
  }
  UNLOCK (&file->lock);

  if (!offset_expected)
    {
      wb_process_queue (process_frame, file, 1);
    }

  wb_enqueue (file, frame, vector, count, offset);
  wb_process_queue (process_frame, file, 0);

  STACK_DESTROY (process_frame->root);

  return 0;
}


int32_t
wb_readv_cbk (call_frame_t *frame,
              void *cookie,
              xlator_t *this,
              int32_t op_ret,
              int32_t op_errno,
              struct iovec *vector,
              int32_t count,
	      struct stat *stbuf)
{
  wb_local_t *local = NULL;

  local = frame->local;
  if (local->file)
    {
      wb_file_unref (local->file);
    }

  STACK_UNWIND (frame, op_ret, op_errno, vector, count, stbuf);
  return 0;
}


int32_t
wb_readv (call_frame_t *frame,
          xlator_t *this,
          fd_t *fd,
          size_t size,
          off_t offset)
{
  wb_file_t *file = NULL;
  wb_local_t *local = NULL;

  if (!dict_get (fd->ctx, this->name))
    {
      gf_log (this->name, GF_LOG_ERROR, "returning EBADFD");
      STACK_UNWIND (frame, -1, EBADFD, NULL);
      return 0;
    }

  file = data_to_ptr (dict_get (fd->ctx, this->name));

  if (file)
    {
      wb_sync_all (frame, file);
    }

  local = calloc (1, sizeof (*local));
  if (file)
    {
      local->file = wb_file_ref (file);
    }
  frame->local = local;

  STACK_WIND (frame,
              wb_readv_cbk,
              FIRST_CHILD(this),
              FIRST_CHILD(this)->fops->readv,
              fd, size, offset);

  return 0;
}


int32_t
wb_ffr_bg_cbk (call_frame_t *frame,
	       void *cookie,
	       xlator_t *this,
	       int32_t op_ret,
	       int32_t op_errno)
{
  wb_local_t *local = NULL;
  wb_file_t *file = NULL;

  local = frame->local;
  file = local->file;

  if (file->op_ret == -1)
    {
      op_ret = file->op_ret;
      op_errno = file->op_errno;

      file->op_ret = 0;
    }
  
  wb_file_unref (file);
  STACK_DESTROY (frame->root);
  return 0;
}


int32_t
wb_ffr_cbk (call_frame_t *frame,
            void *cookie,
            xlator_t *this,
            int32_t op_ret,
            int32_t op_errno)
{
  wb_local_t *local = NULL;
  wb_file_t *file = NULL;

  local = frame->local;
  file = local->file;

  if (file->op_ret == -1)
    {
      op_ret = file->op_ret;
      op_errno = file->op_errno;

      file->op_ret = 0;
    }

  wb_file_unref (file);
  STACK_UNWIND (frame, op_ret, op_errno);
  return 0;
}


int32_t
wb_flush (call_frame_t *frame,
          xlator_t *this,
          fd_t *fd)
{
  wb_conf_t *conf = NULL;
  wb_file_t *file = NULL;
  call_frame_t *flush_frame = NULL;
  wb_local_t *local = NULL;

  conf = this->private;

  if (!dict_get (fd->ctx, this->name))
    {
      gf_log (this->name, GF_LOG_ERROR, "returning EBADFD");
      STACK_UNWIND (frame, -1, EBADFD);
      return 0;
    }

  file = data_to_ptr (dict_get (fd->ctx, this->name));
  local = calloc (1, sizeof (*local));

  if (file)
    {
      local->file = wb_file_ref (file);
    }

  if (&file->request != file->request.next)
    {
      gf_log (this->name, GF_LOG_DEBUG,
	      "request queue is not empty, it has to be synced");
    }

  if (conf->flush_behind && (!file->disabled)) {
    flush_frame = copy_frame (frame);     
    STACK_UNWIND (frame, file->op_ret, file->op_errno); // liar! liar! :O

    flush_frame->local = local;

    wb_sync_all (flush_frame, file);

    STACK_WIND (flush_frame,
                wb_ffr_bg_cbk,
                FIRST_CHILD(this),
                FIRST_CHILD(this)->fops->flush,
                fd);
  } else {
    wb_sync_all (frame, file);

    frame->local = local;
    STACK_WIND (frame,
		wb_ffr_cbk,
		FIRST_CHILD(this),
		FIRST_CHILD(this)->fops->flush,
		fd);
  }

  return 0;
}


int32_t
wb_fsync (call_frame_t *frame,
          xlator_t *this,
          fd_t *fd,
          int32_t datasync)
{
  wb_file_t *file = NULL;
  wb_local_t *local = NULL;

  if (!dict_get (fd->ctx, this->name))
    {
      gf_log (this->name, GF_LOG_ERROR, "returning EBADFD");
      STACK_UNWIND (frame, -1, EBADFD);
      return 0;
    }

  file = data_to_ptr (dict_get (fd->ctx, this->name));

  if (file)
    {
      wb_sync_all (frame, file);
    }

  local = calloc (1, sizeof (*local));
  if (file)
    {
      local->file = wb_file_ref (file);
    }

  frame->local = local;

  STACK_WIND (frame,
              wb_ffr_cbk,
              FIRST_CHILD(this),
              FIRST_CHILD(this)->fops->fsync,
              fd, datasync);
  return 0;
}


int32_t
wb_close (call_frame_t *frame,
          xlator_t *this,
          fd_t *fd)
{
  wb_file_t *file = NULL;
  wb_local_t *local = NULL;

  if (!dict_get (fd->ctx, this->name))
    {
      gf_log (this->name, GF_LOG_ERROR, "returning EBADFD");
      STACK_UNWIND (frame, -1, EBADFD);
      return 0;
    }

  file = data_to_ptr (dict_get (fd->ctx, this->name));
  dict_del (fd->ctx, this->name);

  if (file)
    {
      if (&file->request != file->request.next)
	{
	  gf_log (this->name, GF_LOG_DEBUG,
		  "request queue is not empty, it has to be synced");
	}

      wb_sync_all (frame, file);
    }

  local = calloc (1, sizeof (*local));
  if (file)
    {
      local->file = wb_file_ref (file);
    }
  frame->local = local;

  wb_file_unref (file);

  STACK_WIND (frame,
	      wb_ffr_cbk,
	      FIRST_CHILD(this),
              FIRST_CHILD(this)->fops->close,
              fd);
  return 0;
}


int32_t 
init (xlator_t *this)
{
  dict_t *options = this->options;
  wb_conf_t *conf = NULL;
  char *aggregate_size_string = NULL;
  char *window_size_string = NULL;
  
  if (!this->children || this->children->next)
    {
      gf_log (this->name, GF_LOG_ERROR,
	      "FATAL: write-behind (%s) not configured with exactly one child",
	      this->name);
      return -1;
    }

  conf = calloc (1, sizeof (*conf));

  conf->aggregate_size = 0;
  aggregate_size_string = data_to_str (dict_get (options, 
						 "aggregate-size"));
  if (aggregate_size_string)
    {
      if (gf_string2bytesize (aggregate_size_string, &conf->aggregate_size) != 0)
	{
	  gf_log ("write-behind", 
		  GF_LOG_ERROR, 
		  "invalid number format \"%s\" of \"option aggregate-size\"", 
		  aggregate_size_string);
	  return -1;
	}
    }

  gf_log (this->name,
	  GF_LOG_DEBUG,
	  "using aggregate-size = %d", conf->aggregate_size);
  
  conf->window_size = 0;
  window_size_string = data_to_str (dict_get (options, 
					      "window-size"));
  if (window_size_string)
    {
      if (gf_string2bytesize (window_size_string, &conf->window_size) != 0)
	{
	  gf_log (this->name, GF_LOG_ERROR, 
		  "invalid number format \"%s\" of \"option window-size\"", 
		  window_size_string);
	  FREE (conf);
	  return -1;
	}
    }

  if (!conf->window_size)
    {
      gf_log (this->name, GF_LOG_WARNING,
	      "Setting window-size to be equal to aggregate-size(%"PRId64")",
	      conf->aggregate_size);
      conf->window_size = conf->aggregate_size;
    }

  if (conf->window_size < conf->aggregate_size)
    {
      gf_log (this->name, GF_LOG_ERROR,
	      "aggregate-size(%"PRId64") cannot be more than window-size"
	      "(%"PRId64")", conf->window_size, conf->aggregate_size);
      FREE (conf);
      return -1;
    }


  conf->flush_behind = 0;
  
  if (dict_get (options, "flush-behind"))
    {
      if ((!strcasecmp (data_to_str (dict_get (options, "flush-behind")),
			"on")) ||
	  (!strcasecmp (data_to_str (dict_get (options, "flush-behind")),
			"yes"))) {
	gf_log (this->name, GF_LOG_DEBUG,
		"enabling flush-behind");
	conf->flush_behind = 1;
      }
    }

  this->private = conf;
  return 0;
}


void
fini (xlator_t *this)
{
  wb_conf_t *conf = this->private;

  FREE (conf);
  return;
}


struct xlator_fops fops = {
  .writev      = wb_writev,
  .open        = wb_open,
  .create      = wb_create,
  .readv       = wb_readv,
  .flush       = wb_flush,
  .fsync       = wb_fsync,
  .close       = wb_close,
  .stat        = wb_stat,
  .fstat       = wb_fstat,
  .truncate    = wb_truncate,
  .ftruncate   = wb_ftruncate,
  .utimens     = wb_utimens,
};

struct xlator_mops mops = {
};
